#include <errno.h>
#include <pthread.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/wait.h>
#include <time.h>
#include <unistd.h>


static pthread_mutex_t sMutex = PTHREAD_MUTEX_INITIALIZER;
static volatile int* sSync;


static void
require_errno(bool condition, const char* message,...)
{
	if (condition)
		return;

	va_list args;
	va_start(args, message);
	vfprintf(stderr, message, args);
	va_end(args);

	fprintf(stderr, ": %s\n", strerror(errno));
	exit(1);
}


static void
require_no_error(int error, const char* message,...)
{
	if (error >= 0)
		return;

	va_list args;
	va_start(args, message);
	vfprintf(stderr, message, args);
	va_end(args);

	fprintf(stderr, ": %s\n", strerror(error));
	exit(1);
}


static void
sync_threads(volatile int* syncMe, volatile int* syncOther, int value)
{
	bool meFirst = syncMe < syncOther;
	if (meFirst)
		*syncMe = value;

	while (*syncOther != value) {
	}

	if (!meFirst)
		*syncMe = value;
}


static void*
locker_thread(void* arg)
{
	volatile int* syncMe = (volatile int*)arg;
	volatile int* syncOther = syncMe == sSync ? sSync + 1 : sSync;

	for (;;) {
		sync_threads(syncMe, syncOther, 1);

		timespec timeout;
		require_errno(clock_gettime(CLOCK_REALTIME, &timeout) == 0,
			"failed to get time");
		timeout.tv_nsec += 10000;
		if (timeout.tv_nsec >= 1000000000) {
			timeout.tv_nsec -= 1000000000;
			timeout.tv_sec += 1;
		}

		pthread_mutex_timedlock(&sMutex, &timeout);

		sync_threads(syncMe, syncOther, 2);
	}

	return NULL;
}


int
main()
{
	sSync = new int[2];
	sSync[0] = sSync[1] = 0;

	require_no_error(pthread_mutex_lock(&sMutex), "failed to lock mutex");

	// start the threads
	pthread_t thread;
	require_no_error(pthread_create(&thread, NULL, &locker_thread,
		(void*)sSync), "failed to create thread 1");
	require_no_error(pthread_create(&thread, NULL, &locker_thread,
		(void*)(sSync + 1)), "failed to create thread 1");

	// wait for the threads to be ready
	while (sSync[1] != 1) {
	}

	// fork
	pid_t pid = fork();
	require_errno(pid >= 0, "fork() failed");
	if (pid == 0)
		usleep(100000);
	else
		waitpid(pid, NULL, 0);

	return 0;
}
